/*	
	Yankee !

	variation autour d'un GNG-U ;
	
	d'aprs "Growing Neural Gas, experiments with GNG,
	GNG with utility and supervised GNG."
	de Jim Holmstrm.
	
	Version / 22 mai 2010.
*/

// ----------------------------------------------------------------------------------------------------------------

#include "ext.h"							
#include "ext_obex.h"
#include "ext_common.h"
#include "time.h"	
#include "stdlib.h"
#include "math.h"
#include "ext_systhread.h"
#include "limits.h"

#define TAILLE_VECTEUR_MAX	256
#define	TAILLE_GRAF_MAX		100
#define	TAILLE_LCD_MAX		800
#define	TAILLE_LCD_MIN		100

#define DEF_VALEUR_MAX		127
#define	DEF_TAILLE_VECTEUR	4
#define DEF_TAILLE_MAX		20 
#define DEF_LCD				200
#define DEF_MUTATION		50
#define DEF_MU				5
#define DEF_LAMBDA			2
#define DEF_ALPHA			0.5
#define DEF_BETA			0.25
#define DEF_EPSILON_UN		0.3
#define DEF_EPSILON_DEUX	0.15
#define DEF_KAPPA			3.

// ----------------------------------------------------------------------------------------------------------------

typedef struct _arc {
	long		destination ;
	long		age ;
	} t_arc ;
	
typedef struct _noeud {
	double			erreur ;
	double			utile ;
	double			*valeurs ;
	t_linklist		*list_arcs ;
	} t_noeud ;
	
// ----------------------------------------------------------------------------------------------------------------

typedef struct _yankee {
	t_object			ob ;	
	t_hashtab			*graf ;
	t_linklist			*temp_link ;
	t_atom				*resultat ;
	double				*dump ;
	long				blok ;
	long				dump_k ;
	long				connect ;
	long				nom ;
	long				compteur ;
	long				position ;
	long				lire ;
	long				mutation ;
	long				taille_max ;
	long				taille_graf ;
	long				taille_vecteur ;
	long				valeur_max ;
	long				lcd ;
	long				gagnant1 ;
	long				gagnant2 ;
	long				cumul1 ;
	long				cumul2 ;
	long				perdant ;
	long				mu ;
	long				lambda ;
	double				epsilon1 ;
	double				epsilon2 ;
	double				alpha ;
	double				beta ;
	double				kappa ;
	double				dist1 ;
	double				dist2 ;
	double				erreur1 ;
	double				erreur2 ;
	double				utile ;
	double				*vecteur ;
	void				*m_qelem ;
	void				*m_clock ;
	void				*m_out1 ;
	void				*m_out2 ;
	t_systhread_mutex   c_mutex ;
	} t_yankee ;
	
// ----------------------------------------------------------------------------------------------------------------	

void		*yankee_new					(t_symbol *s, long argc, t_atom *argv) ;
void		yankee_free					(t_yankee *x) ;
void		yankee_assist				(t_yankee *x, void *b, long m, long a, char *s) ;

void		yankee_clear				(t_yankee *x) ;
void		yankee_learn				(t_yankee *x, t_symbol *s, long argc, t_atom *argv) ;
void		yankee_list					(t_yankee *x, t_symbol *s, long argc, t_atom *argv) ;
void		yankee_int					(t_yankee *x, long n) ;
void		yankee_bang					(t_yankee *x) ;

void		graf_initialiser			(t_yankee *x) ;
void		graf_clear					(t_yankee *x) ;
void		graf_deplacer				(t_yankee *x, long n, double f) ;
void		graf_evaluer				(t_yankee *x) ;
void		graf_connecter				(t_yankee *x, long m, long n) ;
void		graf_deconnecter			(t_yankee *x) ;
long		graf_reproduire				(t_yankee *x, long m, long n) ;
void		graf_tuer					(t_yankee *x) ;

void		clear_hashfun				(t_hashtab_entry *e, void *myArg) ;
void		gagner_hashfun				(t_hashtab_entry *e, t_yankee *x) ;
void		evaluer_hashfun				(t_hashtab_entry *e, t_yankee *x) ;
void		vieillir_hashfun			(t_hashtab_entry *e, t_yankee *x) ;
void		mourir_hashfun				(t_hashtab_entry *e, t_yankee *x) ;
void		oublier_hashfun				(t_hashtab_entry *e, t_yankee *x) ;
void		lire_hashfun				(t_hashtab_entry *e, t_yankee *x) ;

void		deplacer_linkfun			(t_object *p, t_yankee *x) ;
void		evaluer_linkfun				(t_object *p, t_yankee *x) ;
void		vieillir_linkfun			(t_object *p, t_yankee *x) ;
void		vieillir_gagnant1_linkfun	(t_object *p, t_yankee *x) ;
void		vieillir_gagnant2_linkfun	(t_object *p, t_yankee *x) ;
void		deconnecter_nouveau_linkfun	(t_object *p, t_yankee *x) ;
void		deconnecter_perdant_linkfun (t_object *p, t_yankee *x) ;
void		tuer_linkfun				(t_object *p, t_yankee *x) ;

void		yankee_dumpmax				(t_yankee *x) ;
void		yankee_task_slow			(t_yankee *x) ;
void		yankee_dumplcd				(t_yankee *x) ;
void		yankee_qtask_lcd			(t_yankee *x) ;

void		dump_hashfun				(t_hashtab_entry *e, t_yankee *x) ;
void		dump_lcd_hashfun			(t_hashtab_entry *e, t_yankee *x) ;

t_max_err	yankee_set_alpha			(t_yankee *x, t_object *attr, long argc, t_atom *argv) ;
t_max_err	yankee_get_alpha			(t_yankee *x, t_object *attr, long *argc, t_atom **argv) ;
t_max_err	yankee_set_beta				(t_yankee *x, t_object *attr, long argc, t_atom *argv) ;
t_max_err	yankee_get_beta				(t_yankee *x, t_object *attr, long *argc, t_atom **argv) ;
t_max_err	yankee_set_epsilon1			(t_yankee *x, t_object *attr, long argc, t_atom *argv) ;
t_max_err	yankee_get_epsilon1			(t_yankee *x, t_object *attr, long *argc, t_atom **argv) ;
t_max_err	yankee_set_epsilon2			(t_yankee *x, t_object *attr, long argc, t_atom *argv) ;
t_max_err	yankee_get_epsilon2			(t_yankee *x, t_object *attr, long *argc, t_atom **argv) ;
t_max_err	yankee_set_kappa			(t_yankee *x, t_object *attr, long argc, t_atom *argv) ;
t_max_err	yankee_get_kappa			(t_yankee *x, t_object *attr, long *argc, t_atom **argv) ;
void		yankee_task					(t_yankee *x) ;

//void		check_hashfun				(t_hashtab_entry *e, t_yankee *x) ;
//void		check_linkfun				(t_object *p, t_yankee *x) ;

// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------

static t_class *yankee_class ;

int main (void)
	{	
		t_class *c = NULL ;
		
		c = class_new ("yankee", (method)yankee_new, (method)yankee_free, (long)sizeof(t_yankee), 0L, A_GIMME, 0) ;

		class_addmethod (c, (method)yankee_assist,		"assist", A_CANT, 0) ; 
		class_addmethod (c, (method)yankee_bang,		"bang", 0) ;
		class_addmethod (c, (method)yankee_clear,		"clear", 0) ;
		class_addmethod (c, (method)yankee_dumpmax,		"dump_max", 0) ;
		class_addmethod (c, (method)yankee_dumplcd,		"dump_lcd", 0) ;
		class_addmethod (c, (method)yankee_learn,		"learn", A_GIMME, 0) ; 
		class_addmethod (c, (method)yankee_list,		"list", A_GIMME, 0) ;
		class_addmethod (c, (method)yankee_list,		"change", A_GIMME, 0) ;
		class_addmethod (c, (method)yankee_int,			"int", A_LONG, 0) ;
		
		class_register (CLASS_BOX, c) ; 
		
		CLASS_ATTR_LONG			(c, "lcd",			ATTR_FLAGS_NONE, t_yankee, lcd) ;
		CLASS_ATTR_LONG			(c, "mutation",		ATTR_FLAGS_NONE, t_yankee, mutation) ;
		CLASS_ATTR_LONG			(c, "mu",			ATTR_FLAGS_NONE, t_yankee, mu) ;
		CLASS_ATTR_LONG			(c, "lambda",		ATTR_FLAGS_NONE, t_yankee, lambda) ;	
		CLASS_ATTR_FLOAT		(c, "alpha",		ATTR_FLAGS_NONE, t_yankee, alpha) ;
		CLASS_ATTR_FLOAT		(c, "beta",			ATTR_FLAGS_NONE, t_yankee, beta) ;
		CLASS_ATTR_FLOAT		(c, "epsilon1",		ATTR_FLAGS_NONE, t_yankee, epsilon1) ;
		CLASS_ATTR_FLOAT		(c, "epsilon2",		ATTR_FLAGS_NONE, t_yankee, epsilon2) ;
		CLASS_ATTR_FLOAT		(c, "kappa",		ATTR_FLAGS_NONE, t_yankee, kappa) ;
		
		CLASS_ATTR_SAVE			(c, "lcd",			ATTR_FLAGS_NONE) ;
		CLASS_ATTR_SAVE			(c, "mutation",		ATTR_FLAGS_NONE) ;
		CLASS_ATTR_SAVE			(c, "mu",			ATTR_FLAGS_NONE) ;
		CLASS_ATTR_SAVE			(c, "lambda",		ATTR_FLAGS_NONE) ;
		CLASS_ATTR_SAVE			(c, "alpha",		ATTR_FLAGS_NONE) ;
		CLASS_ATTR_SAVE			(c, "beta",			ATTR_FLAGS_NONE) ;
		CLASS_ATTR_SAVE			(c, "epsilon1",		ATTR_FLAGS_NONE) ;
		CLASS_ATTR_SAVE			(c, "epsilon2",		ATTR_FLAGS_NONE) ;
		CLASS_ATTR_SAVE			(c, "kappa",		ATTR_FLAGS_NONE) ;
		
		CLASS_ATTR_FILTER_CLIP	(c, "mutation",		0, 100) ;
		CLASS_ATTR_FILTER_CLIP	(c, "mu",			1, 100) ;
		CLASS_ATTR_FILTER_CLIP	(c, "lambda",		1, 100) ;
		CLASS_ATTR_FILTER_CLIP	(c, "alpha",		0.01, 1.) ;
		CLASS_ATTR_FILTER_CLIP	(c, "beta",			0.01, 1.) ;
		CLASS_ATTR_FILTER_CLIP	(c, "epsilon1",		0.01, 1.) ;
		CLASS_ATTR_FILTER_CLIP	(c, "epsilon2",		0.01, 1.) ;
		CLASS_ATTR_FILTER_CLIP	(c, "kappa",		0, 100) ;
		CLASS_ATTR_FILTER_CLIP	(c, "lcd",			TAILLE_LCD_MIN, TAILLE_LCD_MAX) ;
		
		CLASS_ATTR_ORDER		(c, "mu",			ATTR_FLAGS_NONE, "1") ;
		CLASS_ATTR_ORDER		(c, "lambda",		ATTR_FLAGS_NONE, "2") ;
		CLASS_ATTR_ORDER		(c, "alpha",		ATTR_FLAGS_NONE, "3") ;
		CLASS_ATTR_ORDER		(c, "beta",			ATTR_FLAGS_NONE, "4") ;
		CLASS_ATTR_ORDER		(c, "epsilon1",		ATTR_FLAGS_NONE, "5") ;
		CLASS_ATTR_ORDER		(c, "epsilon2",		ATTR_FLAGS_NONE, "6") ;
		CLASS_ATTR_ORDER		(c, "kappa",		ATTR_FLAGS_NONE, "7") ;
		CLASS_ATTR_ORDER		(c, "lcd",			ATTR_FLAGS_NONE, "8") ;
		CLASS_ATTR_ORDER		(c, "mutation",		ATTR_FLAGS_NONE, "9") ;
		
		CLASS_ATTR_ACCESSORS	(c, "alpha",		yankee_get_alpha, yankee_set_alpha) ;
		CLASS_ATTR_ACCESSORS	(c, "beta",			yankee_get_beta, yankee_set_beta) ;
		CLASS_ATTR_ACCESSORS	(c, "epsilon1",		yankee_get_epsilon1, yankee_set_epsilon1) ;
		CLASS_ATTR_ACCESSORS	(c, "epsilon2",		yankee_get_epsilon2, yankee_set_epsilon2) ;
		CLASS_ATTR_ACCESSORS	(c, "kappa",		yankee_get_kappa, yankee_set_kappa) ;
				
		yankee_class = c ;
		
		return MAX_ERR_NONE ;
	}

// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------

void *yankee_new (t_symbol *s, long argc, t_atom *argv)
	{
		t_yankee *x = NULL ;
		
		if (x = (t_yankee *)object_alloc (yankee_class)) 
			{
				long ac = 0 ;
				char alloc ;
				
				srand (time(NULL)) ;
				
				x->blok				= 0 ;
				x->taille_graf		= 0 ;
				x->dump_k			= 0 ;
				x->connect			= 0 ;
				x->nom				= 0 ;
				x->compteur			= 0 ;
				x->position			= 0 ;
				x->lire				= 0 ;
				x->gagnant1			= 0 ;
				x->gagnant2			= 0 ;
				x->cumul1			= 0 ;
				x->cumul2			= 0 ;
				x->perdant			= 0 ;
				x->dist1			= 0.;
				x->dist2			= 0.;
				x->erreur1			= 0.;
				x->erreur2			= 0.;
				x->utile			= 0.;
				x->taille_vecteur	= DEF_TAILLE_VECTEUR ;
				x->taille_max		= DEF_TAILLE_MAX ;
				x->valeur_max		= DEF_VALEUR_MAX ;
				x->lcd				= DEF_LCD ;
				x->mutation			= DEF_MUTATION ;
				x->alpha			= DEF_ALPHA ;
				x->beta				= DEF_BETA ;
				x->mu				= DEF_MU ;
				x->lambda			= DEF_LAMBDA ;
				x->epsilon1			= DEF_EPSILON_UN ;
				x->epsilon2			= DEF_EPSILON_DEUX ;
				x->kappa			= DEF_KAPPA ;
				x->graf				= NULL ;
				x->resultat			= NULL ; 
				x->temp_link		= NULL ;
				x->dump				= NULL ;

				if (argc)
					{
						if (atom_gettype (argv) == A_LONG)
							{
								long k = atom_getlong (argv) ;
								
								if (k > 1 && k <= TAILLE_GRAF_MAX)
									{
										x->taille_max = k ;
									}
								
								if (argc > 1)
									{
										if (atom_gettype (argv + 1) == A_LONG)
											{
												long w = atom_getlong (argv + 1) ;
												
												if (w > 1 && w <= LONG_MAX)
													{
														x->valeur_max = w ;
													}
											}
									}
							}
					}
				
				attr_args_process (x, argc, argv) ;
				
				x->vecteur	= (double *)sysmem_newptrclear (sizeof(double) * TAILLE_VECTEUR_MAX) ;
				x->dump		= (double *)sysmem_newptrclear (sizeof(double) * TAILLE_VECTEUR_MAX * x->taille_max) ;
				
				atom_alloc_array (TAILLE_VECTEUR_MAX, &ac, &x->resultat, &alloc) ;
			
				if (x->graf = hashtab_new (0)) 
					{
						hashtab_flags (x->graf, OBJ_FLAG_MEMORY) ;
					}
					
				graf_initialiser (x) ;
				
				x->m_out2	= outlet_new ((t_object *)x, NULL) ;
				x->m_out1	= listout ((t_object *)x) ;
				
				x->m_qelem  = qelem_new ((t_object *)x, (method)yankee_qtask_lcd) ;
				x->m_clock	= clock_new ((t_object *)x, (method)yankee_task);
				
				systhread_mutex_new (&x->c_mutex, SYSTHREAD_MUTEX_NORMAL) ;
			}
				
		return x ;
	}
	
void yankee_free (t_yankee *x)
	{ 
		if (x->resultat)
			{
				sysmem_freeptr (x->resultat) ;
			}
			
		if (x->vecteur)
			{
				sysmem_freeptr (x->vecteur) ;
			}
		
		if (x->dump)
			{
				sysmem_freeptr (x->dump) ;
			}

		graf_clear (x) ;

		if (x->graf)
			{
				object_free (x->graf) ;
			}
		
		qelem_unset (x->m_qelem) ;
		qelem_free	(x->m_qelem) ;
		
		clock_unset (x->m_clock) ;
		object_free (x->m_clock) ;
		
		systhread_mutex_free (x->c_mutex) ;
	}

void yankee_assist (t_yankee *x, void *b, long m, long a, char *s)
	{
		if (m == ASSIST_INLET) 
			{ 
				sprintf (s, "(list) (bang) (int) (symbol) clear, learn, change, dump_max, dump_lcd") ;
			} 
		else 
			{	
				if (a == 0 )
					{
						sprintf (s, "(list)") ;	
					}
				else
					{
						sprintf (s, "(lcd) dump") ;
					}
			}
	}

// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------

void yankee_clear (t_yankee *x) 
	{
		if (!systhread_mutex_trylock (x->c_mutex))
			{
				graf_clear			(x) ;
				graf_initialiser	(x) ;
				
				systhread_mutex_unlock (x->c_mutex) ;
			}
	}
	
void yankee_learn (t_yankee *x, t_symbol *s, long argc, t_atom *argv)
	{
		if (!systhread_mutex_trylock (x->c_mutex))
			{
				if (argc != x->taille_vecteur && argc <= TAILLE_VECTEUR_MAX)  
					{
						x->blok	= 1 ;
						
						x->taille_vecteur = argc ;
							
						graf_clear			(x) ;			
						graf_initialiser	(x) ;
						
						clock_fdelay (x->m_clock, 1000.) ;
					}
					
				if (x->vecteur && x->graf && argc == x->taille_vecteur)
					{
						long	i ;
						t_noeud *nptr = NULL ;
						
						for (i = 0 ; i < argc ; i ++)
							{
								if (atom_gettype (argv + i) != A_LONG)
									{		
										x->vecteur[i] = rand () % (x->valeur_max + 1) ;
									}
								else
									{
										x->vecteur[i] = CLIP ((double)atom_getlong (argv + i), 0., (double)x->valeur_max) ;
									}
							}
							
						hashtab_funall (x->graf, (method)gagner_hashfun, x) ;
						
						if (hashtab_lookup (x->graf, (t_symbol *)x->gagnant1, (t_object **)&nptr) == MAX_ERR_NONE)
							{
								long j ;
								
								nptr->erreur += x->dist1 ;
								nptr->utile  += x->dist2 - x->dist1 ;
								
								for (j = 0 ; j < x->taille_vecteur ; j++)
									{
										if (nptr->valeurs)
											{
												nptr->valeurs[j] += x->epsilon1 * (x->vecteur[j] - nptr->valeurs[j]) ;
											}
									}
								
								if (nptr->list_arcs)
									{
										linklist_funall (nptr->list_arcs, (method)deplacer_linkfun, x) ;
									}
							}
						
						hashtab_funall (x->graf, (method)vieillir_hashfun, x) ;
						
						if (!x->connect)
							{
								graf_connecter (x, x->gagnant1, x->gagnant2) ;
								graf_connecter (x, x->gagnant2, x->gagnant1) ;
								
								x->connect = 0 ;
							}
							
						hashtab_funall (x->graf, (method)mourir_hashfun, x) ;
						
						graf_evaluer (x) ;
						
						if (!(x->compteur % x->lambda))
							{
								if ((x->erreur1 / x->utile) > x->kappa && x->taille_graf == x->taille_max)
									{
										graf_tuer (x) ;
									}
								
								if (x->taille_graf < x->taille_max)
									{
										long k = graf_reproduire (x, x->cumul1, x->cumul2) ;		
														
										graf_connecter		(x, x->cumul1, k) ;
										graf_connecter		(x, x->cumul2, k) ;						
										graf_deconnecter	(x) ;	
															
										hashtab_funall (x->graf, (method)oublier_hashfun, x) ;
									}
							}
							
						x->gagnant1	= 0 ;
						x->gagnant2	= 0 ;
						x->cumul1	= 0 ;
						x->cumul2	= 0 ;
						x->perdant	= 0 ;
						
						x->compteur ++ ;
						
						systhread_mutex_unlock (x->c_mutex) ;
					} 
			}
	}
		
void yankee_list (t_yankee *x, t_symbol *s, long argc, t_atom *argv)
	{
		if (x->graf && x->resultat)
			{
				t_atom	temp[TAILLE_VECTEUR_MAX] ;
				t_atom	*aptr = x->resultat ;
				
				systhread_mutex_lock (x->c_mutex) ;
		
				if (x->taille_graf > 2)
					{
						long i ;
						
						x->lire = rand () % x->taille_graf ;
						x->position = 0 ;
		
						hashtab_funall (x->graf, (method)lire_hashfun, x) ;
						
						for (i = 0 ; i < argc ; i++)
							{
								long h = rand () % 100 ;
								
								if (h < x->mutation)
									{
										atom_setlong (temp + i, atom_getlong (aptr + (i % x->taille_vecteur))) ;
									}
								else
									{
										atom_setlong (temp + i, atom_getlong (argv + i)) ;
									}
							}
					}
				else
					{
						long i ;
						
						for (i = 0 ; i < argc ; i++)
							{
								atom_setlong (temp + i, atom_getlong (argv + i)) ;
							}
					}
				
				systhread_mutex_unlock (x->c_mutex) ;
				
				outlet_list (x->m_out1, NULL, (short)argc, temp) ;
			}
	}

void yankee_int (t_yankee *x, long n)
	{
		if (x->graf && x->resultat)
			{
				systhread_mutex_lock (x->c_mutex) ;
		
				if (x->taille_graf  && n >= 1 && n <= x->taille_graf)
					{
						x->lire = n - 1 ;
						x->position = 0 ;
		
						hashtab_funall (x->graf, (method)lire_hashfun, x) ;
					}
				
				systhread_mutex_unlock (x->c_mutex) ;
				
				if (n >= 1 && n <= x->taille_graf)
					{
						t_atom *aptr = x->resultat ;
						
						outlet_list (x->m_out1, NULL, (short)x->taille_vecteur, aptr) ;
					}
			}	
	}

void yankee_bang (t_yankee *x)
	{
		if (x->graf && x->resultat)
			{
				t_atom *aptr = x->resultat ;
				
				systhread_mutex_lock (x->c_mutex) ;
		
				if (x->taille_graf)
					{
						x->lire = rand () % x->taille_graf ;
						x->position = 0 ;
		
						hashtab_funall (x->graf, (method)lire_hashfun, x) ;
					}
				
				systhread_mutex_unlock (x->c_mutex) ;
				
				outlet_list (x->m_out1, NULL, (short)x->taille_vecteur, aptr) ;
			}
	}
		
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------	
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------

void graf_initialiser (t_yankee *x)
	{
		if (x->graf)
			{
				t_noeud *nptr1 = NULL ;
				t_noeud *nptr2 = NULL ;
				
				if (nptr1 = (t_noeud *)sysmem_newptr (sizeof(t_noeud)))
					{
						if (nptr1->valeurs = (double *)sysmem_newptrclear (sizeof(double) * TAILLE_VECTEUR_MAX))
							{
								long i ;
								
								for (i = 0 ; i < x->taille_vecteur ; i++)
									{
										nptr1->valeurs[i] = (double)(rand () % (x->valeur_max + 1)) ;
									}
							}
							
						nptr1->erreur	= 0. ;
						nptr1->utile	= 1. ;
						
						if (nptr1->list_arcs = linklist_new ())
							{
								t_arc *aptr = NULL ;
								
								linklist_flags (nptr1->list_arcs, OBJ_FLAG_MEMORY) ;
								
								if (aptr = (t_arc *)sysmem_newptr (sizeof(t_arc)))
									{
										aptr->destination	= 2 ;
										aptr->age			= 0 ;
									}

								linklist_append (nptr1->list_arcs, aptr) ;
							}
					}
				
				if (nptr2 = (t_noeud *)sysmem_newptr (sizeof(t_noeud)))
					{
						if (nptr2->valeurs = (double *)sysmem_newptrclear (sizeof(double) * TAILLE_VECTEUR_MAX))
							{
								long i ;
								
								for (i = 0 ; i < x->taille_vecteur ; i++)
									{
										nptr2->valeurs[i] = (double)(rand () % (x->valeur_max + 1)) ;
									}
							}
							
						nptr2->erreur	= 0. ;
						nptr2->utile	= 1. ;
						
						if (nptr2->list_arcs = linklist_new ())
							{
								t_arc *aptr = NULL ;
								
								linklist_flags (nptr2->list_arcs, OBJ_FLAG_MEMORY) ;
								
								if (aptr = (t_arc *)sysmem_newptr (sizeof(t_arc)))
									{
										aptr->destination	= 1 ;
										aptr->age			= 0 ;
									}

								linklist_append (nptr2->list_arcs, aptr) ;
							}
					}
					
				hashtab_store (x->graf, (t_symbol *)1, (t_object *)nptr1) ;
				hashtab_store (x->graf, (t_symbol *)2, (t_object *)nptr2) ;
				
				x->taille_graf	= 2 ;
				x->nom			= 3 ;
				x->compteur		= 0 ;
			}
	}		

void graf_clear (t_yankee *x)
	{
		if (x->graf)
			{
				hashtab_funall (x->graf, (method)clear_hashfun, NULL) ;
				hashtab_clear  (x->graf) ;
			}
	}

void graf_evaluer (t_yankee *x) 
	{
		t_noeud *nptr = NULL ;
		
		hashtab_funall (x->graf, (method)evaluer_hashfun, x) ;
		
		if (hashtab_lookup (x->graf, (t_symbol *)x->cumul1, (t_object **)&nptr) == MAX_ERR_NONE)
			{
				if (nptr->list_arcs)
					{
						linklist_funall (nptr->list_arcs, (method)evaluer_linkfun, x) ;
					}
			}
	}
	
void graf_deplacer (t_yankee *x, long n, double f) 
	{
		t_noeud *nptr = NULL ;
		
		if (hashtab_lookup (x->graf, (t_symbol *)n, (t_object **)&nptr) == MAX_ERR_NONE)
			{
				long i ;
				
				for (i = 0 ; i < x->taille_vecteur ; i++)
					{
						if (nptr->valeurs)
							{
								nptr->valeurs[i] += f * (x->vecteur[i] - nptr->valeurs[i]) ;
							}
					}
			}
	}

void graf_connecter (t_yankee *x, long m, long n)
	{
		t_noeud *nptr = NULL ;
		
		if (hashtab_lookup (x->graf, (t_symbol *)m, (t_object **)&nptr) == MAX_ERR_NONE)
			{
				t_arc *aptr = NULL ;
				
				if (aptr = (t_arc *)sysmem_newptr (sizeof(t_arc)))
					{
						aptr->destination	= n ;
						aptr->age			= 0 ;
					}
				
				if (nptr->list_arcs)
					{
						linklist_append (nptr->list_arcs, aptr) ;
					}
			}
	}

void graf_deconnecter (t_yankee *x)
	{
		t_noeud *nptr = NULL ;
		
		if (hashtab_lookup (x->graf, (t_symbol *)x->cumul1, (t_object **)&nptr) == MAX_ERR_NONE)
			{
				if (x->temp_link = nptr->list_arcs)
					{
						linklist_funall (nptr->list_arcs, (method)deconnecter_nouveau_linkfun, x)  ;
					}
			}
		
		if (hashtab_lookup (x->graf, (t_symbol *)x->cumul2, (t_object **)&nptr) == MAX_ERR_NONE)
			{
				if (x->temp_link = nptr->list_arcs)
					{
						linklist_funall (nptr->list_arcs, (method)deconnecter_nouveau_linkfun, x)  ;
					}
			}
	}

long graf_reproduire (t_yankee *x, long m, long n) 
	{
		long		k = 0 ;
		t_noeud		*nptr1 = NULL ;
		t_noeud		*nptr2 = NULL ;
		t_max_err	err1 = hashtab_lookup (x->graf, (t_symbol *)m, (t_object **)&nptr1) ;
		t_max_err	err2 = hashtab_lookup (x->graf, (t_symbol *)n, (t_object **)&nptr2) ;
	
		if (!err1 && !err2)
			{
				nptr1->erreur = x->alpha * nptr1->erreur ;
				nptr2->erreur = x->alpha * nptr2->erreur ;
				
				if (nptr1->valeurs && nptr2->valeurs)
					{
						t_noeud *nptr = NULL ;
						
						if (nptr = (t_noeud *)sysmem_newptr (sizeof(t_noeud)))
							{
								if (nptr->valeurs = (double *)sysmem_newptrclear (sizeof(double) * TAILLE_VECTEUR_MAX))
									{
										long i ;
										
										for (i = 0 ; i < x->taille_vecteur ; i++)
											{
												nptr->valeurs[i] =  (nptr1->valeurs[i] + nptr2->valeurs[i]) / 2 ;
											}
									}
									
								nptr->erreur	= (nptr1->erreur + nptr2->erreur) / 2 ;
								nptr->utile		= (nptr1->utile + nptr2->utile) / 2 ;
								
								if (nptr->list_arcs = linklist_new ())
									{
										t_arc *aptr1 = NULL ;
										t_arc *aptr2 = NULL ;
										
										linklist_flags (nptr->list_arcs, OBJ_FLAG_MEMORY) ;
										
										if (aptr1 = (t_arc *)sysmem_newptr (sizeof(t_arc)))
											{
												aptr1->destination	= x->cumul1 ;
												aptr1->age			= 0 ;
											}
										
										if (aptr2 = (t_arc *)sysmem_newptr (sizeof(t_arc)))
											{
												aptr2->destination	= x->cumul2 ;
												aptr2->age			= 0 ;
											}

										linklist_append (nptr->list_arcs, aptr1) ;
										linklist_append (nptr->list_arcs, aptr2) ;
									}
							}
						
						k = x->nom ;	
						hashtab_store (x->graf, (t_symbol *)k, (t_object *)nptr) ;
						x->taille_graf ++ ;
						x->nom ++ ;
					}
			}

		return k ;
	}

void graf_tuer (t_yankee *x) 
	{
		t_noeud *nptr = NULL ;
		
		if (hashtab_lookup (x->graf, (t_symbol *)x->perdant, (t_object **)&nptr) == MAX_ERR_NONE)
			{
				if (nptr->list_arcs)
					{
						linklist_funall (nptr->list_arcs, (method)tuer_linkfun, x) ;
					}
					
				if (nptr->valeurs)
					{
						sysmem_freeptr (nptr->valeurs) ;
					}
				
				if (nptr->list_arcs)
					{
						object_free (nptr->list_arcs) ;
					}
										
				hashtab_delete (x->graf, (t_symbol *)x->perdant) ;
				x->taille_graf -- ;
			}
	}
	
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------

void clear_hashfun (t_hashtab_entry *e, void *myArg)
    {
        if (e->key) 
			{
				t_noeud *nptr = NULL ;
				
				if (nptr = (t_noeud *)e->value)
					{
						if (nptr->valeurs)
							{
								sysmem_freeptr (nptr->valeurs) ;
							}
								
						if (nptr->list_arcs)
							{
								linklist_clear (nptr->list_arcs) ;
								object_free (nptr->list_arcs) ;
							}
						
						sysmem_freeptr (nptr) ;
					}
			}
    }

void gagner_hashfun (t_hashtab_entry *e, t_yankee *x)
    {
        if (e->key) 
			{
				t_noeud *nptr = NULL ;
				
				if (nptr = (t_noeud *)e->value)
					{
						double	k  = 0. ;
						double	*p = NULL ;
						double  *q = x->vecteur ;
						
						if (p = nptr->valeurs)
							{
								long i ;
								
								for (i = 0 ; i < x->taille_vecteur ; i++)
									{
										k += pow (p[i] - q[i], 2) ;
									}	
									
								if (!x->gagnant1)
									{
										x->gagnant1 = (long)e->key ;
										x->dist1	= k ;
									}
								else if (!x->gagnant2)
									{
										if (k < x->dist1)
											{
												x->gagnant2 = x->gagnant1 ;
												x->gagnant1 = (long)e->key ;
												x->dist2	= x->dist1 ;
												x->dist1	= k ;
											}
										else
											{	
												x->gagnant2 = (long)e->key ;
												x->dist2	= k ;
											}
									}
								else
									{
										if (k < x->dist1)
											{
												x->gagnant2 = x->gagnant1 ;
												x->dist2	= x->dist1 ;
												x->gagnant1 = (long)e->key ;
												x->dist1	= k ;
											}
										else if (k < x->dist2)
											{
												x->gagnant2 = (long)e->key ;
												x->dist2	= k ;
											}
									}
							}
					}
			}
    }

void evaluer_hashfun (t_hashtab_entry *e, t_yankee *x)
    {
        if (e->key) 
			{
				t_noeud *nptr = NULL ;
				
				if (nptr = (t_noeud *)e->value)
					{
						double k  = nptr->erreur ;
							
						if (!x->cumul1)
							{
								x->cumul1 = (long)e->key ;
								x->erreur1	= k ;
							}
						else
							{
								if (k > x->erreur1)
									{
										x->cumul1 = (long)e->key ;
										x->erreur1	= k ;
									}
							}
					}
			}
    }
	
void vieillir_hashfun (t_hashtab_entry *e, t_yankee *x) 
	{
		t_noeud *nptr = NULL ;
		
		if ((long)e->key == x->gagnant1) 
			{
				if (nptr = (t_noeud *)e->value)
					{
						if (x->temp_link = nptr->list_arcs)
							{
								linklist_funall (nptr->list_arcs, (method)vieillir_gagnant1_linkfun, x) ;
							}
					}
			}
		else if ((long)e->key == x->gagnant2) 
			{
				if (nptr = (t_noeud *)e->value)
					{
						if (x->temp_link = nptr->list_arcs)
							{
								linklist_funall (nptr->list_arcs, (method)vieillir_gagnant2_linkfun, x) ;
							}
					}
			}
		else
			{
				if (nptr = (t_noeud *)e->value)
					{
						if (x->temp_link = nptr->list_arcs)
							{
								linklist_funall (nptr->list_arcs, (method)vieillir_linkfun, x) ;
							}
					}
			}
	}

void mourir_hashfun	(t_hashtab_entry *e, t_yankee *x) 
	{
		 if (e->key) 
			{
				t_noeud *nptr = NULL ;
				
				if (nptr = (t_noeud *)e->value)
					{
						if (nptr->list_arcs)
							{
								long k = linklist_getsize (nptr->list_arcs) ;
								
								if (!k)
									{
										if (nptr->valeurs)
											{
												sysmem_freeptr (nptr->valeurs) ;
											}
										
										if (nptr->list_arcs)
											{
												object_free (nptr->list_arcs) ;
											}
										
										hashtab_delete (x->graf, (t_symbol *)e->key) ;
										x->taille_graf -- ;
									}
								else
									{
											double k = nptr->utile ;
							
											if (!x->perdant)
												{
													x->perdant	= (long)e->key ;
													x->utile	= k ;
												}
											else
												{
													if (k < x->utile)
														{
															x->perdant	= (long)e->key ;
															x->utile	= k ;
														}
												}
									}
							}
					}
			}
	}

void oublier_hashfun (t_hashtab_entry *e, t_yankee *x) 
	{
		 if (e->key) 
			{
				t_noeud *nptr = NULL ;
				
				if (nptr = (t_noeud *)e->value)
					{
						nptr->erreur	= nptr->erreur - (x->beta * nptr->erreur) ;
						nptr->utile		= nptr->utile - (x->beta * nptr->utile) ;
					}
			}
	}

void lire_hashfun (t_hashtab_entry *e, t_yankee *x)
	{
		 if (e->key) 
			{
				t_noeud *nptr = (t_noeud *)e->value ;
				
				if (nptr && (x->position == x->lire))
					{
						double	*ptr = NULL ;
						t_atom  *aptr = x->resultat ;
						
						if (ptr = nptr->valeurs)
							{
								long i ;
								
								for (i = 0 ; i < x->taille_vecteur ; i++)
									{
										atom_setlong (aptr + i, (long)round (ptr[i])) ;
									}
							}
					}
				
				x->position ++ ;
			}
	}

// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------

void deplacer_linkfun (t_object *p, t_yankee *x)
	{
		t_arc *aptr = NULL ;
		
		if (aptr = (t_arc *)p)
			{
				graf_deplacer (x, aptr->destination, x->epsilon2) ;
			}
	}

void evaluer_linkfun (t_object *p, t_yankee *x)
	{
		t_arc *aptr = NULL ;
		
		if (aptr = (t_arc *)p)
			{
				t_noeud *nptr = NULL ;
				
				if (hashtab_lookup (x->graf, (t_symbol *)aptr->destination, (t_object **)&nptr) == MAX_ERR_NONE)
					{
						double k  = nptr->erreur ;
							
						if (!x->cumul2)
							{
								x->cumul2 = aptr->destination ;
								x->erreur2	= k ;
							}
						else
							{
								if (k > x->erreur2)
									{
										x->cumul2 = aptr->destination ;
										x->erreur2	= k ;
									}
							}
					}
			}
	}

void vieillir_linkfun (t_object *p, t_yankee *x)
	{
		t_arc *aptr = NULL ;
		
		if (aptr = (t_arc *)p)
			{
				if (aptr->destination == x->gagnant1)
					{
						aptr->age ++ ;
						
						if (aptr->age > x->mu)
							{
								linklist_deleteindex (x->temp_link, linklist_objptr2index (x->temp_link, aptr)) ;
							}
					}
			}
	}
	
void vieillir_gagnant1_linkfun (t_object *p, t_yankee *x)
	{
		t_arc *aptr = NULL ;
		
		if (aptr = (t_arc *)p)
			{
				if (aptr->destination == x->gagnant2)
					{
						aptr->age = 0 ;
						x->connect = 1 ;
					}
				else
					{
						aptr->age ++ ;
						
						if (aptr->age > x->mu)
							{
								linklist_deleteindex (x->temp_link, linklist_objptr2index (x->temp_link, aptr)) ;
							}
					}
			}
	}

void vieillir_gagnant2_linkfun (t_object *p, t_yankee *x)
	{
		t_arc *aptr = NULL ;
				
		if (aptr = (t_arc *)p)
			{
				if (aptr->destination == x->gagnant1)
					{
						aptr->age = 0 ;
					}		
			}
	}

void deconnecter_nouveau_linkfun (t_object *p, t_yankee *x) 
	{
		t_arc *aptr = NULL ;
		
		if (aptr = (t_arc *)p)
			{
				if (aptr->destination == x->cumul1 || aptr->destination == x->cumul2)
					{
						linklist_deleteindex (x->temp_link, linklist_objptr2index (x->temp_link, aptr)) ;
					}
			}
	}

void deconnecter_perdant_linkfun (t_object *p, t_yankee *x) 
	{
		t_arc *aptr = NULL ;
		
		if (aptr = (t_arc *)p)
			{
				if (aptr->destination == x->perdant)
					{
						linklist_deleteindex (x->temp_link, linklist_objptr2index (x->temp_link, aptr)) ;
					}
			}
	}

void tuer_linkfun (t_object *p, t_yankee *x)
	{
		t_arc *aptr = NULL ;
		
		if (aptr = (t_arc *)p)
			{
				t_noeud *nptr = NULL ;
				
				if (hashtab_lookup (x->graf, (t_symbol *)aptr->destination, (t_object **)&nptr) == MAX_ERR_NONE)
					{
						if (x->temp_link = nptr->list_arcs)
							{
								linklist_funall (nptr->list_arcs, (method)deconnecter_perdant_linkfun, x) ;
							}
					}
			}
	}

// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------

void yankee_dumpmax (t_yankee *x) 
	{
		if (x->graf && x->dump)
			{
				long	i ;
				double	*ptr = x->dump ;
								
				if (!systhread_mutex_trylock (x->c_mutex))
					{
						x->dump_k = 0 ;
						
						hashtab_funall	(x->graf, (method)dump_hashfun, x) ;
						
						systhread_mutex_unlock (x->c_mutex) ;
					}
				
				post ("GNG : %ld / %ld",	x->taille_graf, x->taille_max) ;
				post ("mu : %ld",			x->mu) ;
				post ("lambda : %ld",		x->lambda) ;
				post ("alpha : %.2lf",		x->alpha) ;
				post ("beta : %.2lf",		x->beta) ;
				post ("epsilon : %.2lf",	x->epsilon1) ;
				post ("epsilon : %.2lf",	x->epsilon2) ;
				post ("kappa : %.2lf",		x->kappa) ;
				
				post (" ") ;
				
				for (i = 0 ; i < x->taille_graf ; i++)
					{
						long	j ;
						long	text_max = 100 ;
						long	ac = x->taille_vecteur ;
						t_atom	temp[TAILLE_VECTEUR_MAX] ;
						char	*affichage = NULL ;
						
						for (j = 0 ; j < ac ; j++) 
							{
								atom_setlong (temp + j, (long)round (ptr[j + (i * x->taille_vecteur)])) ;
							}
						
						if (affichage = (char *)sysmem_newptr (sizeof(char) * text_max))
							{
								atom_gettext (ac, temp, &text_max, &affichage, OBEX_UTIL_ATOM_GETTEXT_DEFAULT) ; 
							
								post ("%s", affichage) ;	
								
								sysmem_freeptr (affichage) ;
							}
					}
				
				post (" ") ;
				
				//hashtab_funall	(x->graf, (method)check_hashfun, x) ;
			}
	}
	
void yankee_dumplcd (t_yankee *x) 
	{
		if (x->graf && x->dump && !x->blok)
			{
				if (!systhread_mutex_trylock (x->c_mutex))
					{
						x->dump_k = 0 ;
						
						hashtab_funall (x->graf, (method)dump_hashfun, x) ;
						
						systhread_mutex_unlock (x->c_mutex) ;
					}
				
				qelem_set (x->m_qelem) ;
			}
	}

void yankee_qtask_lcd (t_yankee *x)
	{	
		long	i ;
		long	alpha, beta ;
		t_atom	lcdsize[2] ;
		double	*ptr = x->dump ;

		alpha	= MAX (x->lcd / x->taille_max, 2) ;
		beta	= MAX (x->lcd / x->taille_vecteur, 2) ;
		
		atom_setlong (lcdsize, alpha * x->taille_max) ;
		atom_setlong (lcdsize + 1, beta * x->taille_vecteur) ;
		
		outlet_anything (x->m_out2, gensym("clear"), 0, NULL) ;
		outlet_anything (x->m_out2, gensym("size"), 2, lcdsize) ;
		
		for (i = 0 ; i < x->taille_graf ; i++)
			{
				long j ;
								
				for (j = 0 ; j < x->taille_vecteur ; j++)
					{
						t_atom	argv[5] ;
						long	k = (long)round (((ptr[j + (i * x->taille_vecteur)]) / x->valeur_max) * 128) ;
		
						atom_setlong (&argv[0], i			* alpha) ;
						atom_setlong (&argv[1], j			* beta) ;
						atom_setlong (&argv[2], ((i + 1)	* alpha)) ;
						atom_setlong (&argv[3], ((j + 1)	* beta)) ;
						atom_setlong (&argv[4], CLIP (k, 0, 255)) ;
				
						outlet_anything (x->m_out2, gensym ("paintrect"), 5, argv) ;
					}
			}
	}

void dump_hashfun (t_hashtab_entry *e, t_yankee *x) 
	{
		if (e->key) 
			{
				t_noeud *nptr = NULL ;
				
				if (nptr = (t_noeud *)e->value)
					{
						double *ptr = NULL ;
						
						if (ptr = nptr->valeurs)
							{
								long i ;
								
								for (i = 0 ; i < x->taille_vecteur ; i++)
									{
										x->dump[i + (x->dump_k * x->taille_vecteur)] = ptr[i] ;
									}
							}
						
						x->dump_k ++ ;
					}
			}
	}

// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
	
void yankee_task (t_yankee *x)
	{
		x->blok = 0 ;
	}

t_max_err yankee_set_alpha (t_yankee *x, t_object *attr, long argc, t_atom *argv)
    {
        x->alpha = atom_getfloat (argv) ;

        return MAX_ERR_NONE ;
    }
	
t_max_err yankee_get_alpha (t_yankee *x, t_object *attr, long *argc, t_atom **argv)
    {
        char		alloc ;
		double		k = x->alpha ; 
 
        atom_alloc		(argc, argv, &alloc) ;     
        atom_setfloat	(*argv, k) ;
		
        return MAX_ERR_NONE ;
    }

t_max_err yankee_set_beta (t_yankee *x, t_object *attr, long argc, t_atom *argv)
    {
        x->beta = atom_getfloat (argv) ;

        return MAX_ERR_NONE ;
    }
	
t_max_err yankee_get_beta (t_yankee *x, t_object *attr, long *argc, t_atom **argv)
    {
        char		alloc ;
		double		k = x->beta ; 
 
        atom_alloc		(argc, argv, &alloc) ;     
        atom_setfloat	(*argv, k) ;
		
        return MAX_ERR_NONE ;
    }

t_max_err yankee_set_epsilon1 (t_yankee *x, t_object *attr, long argc, t_atom *argv)
    {
        x->epsilon1 = atom_getfloat (argv) ;

        return MAX_ERR_NONE ;
    }
	
t_max_err yankee_get_epsilon1 (t_yankee *x, t_object *attr, long *argc, t_atom **argv)
    {
        char		alloc ;
		double		k = x->epsilon1 ; 
 
        atom_alloc		(argc, argv, &alloc) ;     
        atom_setfloat	(*argv, k) ;
		
        return MAX_ERR_NONE ;
    }

t_max_err yankee_set_epsilon2 (t_yankee *x, t_object *attr, long argc, t_atom *argv)
    {
        x->epsilon2 = atom_getfloat (argv) ;

        return MAX_ERR_NONE ;
    }
	
t_max_err yankee_get_epsilon2 (t_yankee *x, t_object *attr, long *argc, t_atom **argv)
    {
        char		alloc ;
		double		k = x->epsilon2 ; 
 
        atom_alloc		(argc, argv, &alloc) ;     
        atom_setfloat	(*argv, k) ;
		
        return MAX_ERR_NONE ;
    }

t_max_err yankee_set_kappa (t_yankee *x, t_object *attr, long argc, t_atom *argv)
    {
        x->kappa = atom_getfloat (argv) ;

        return MAX_ERR_NONE ;
    }
	
t_max_err yankee_get_kappa (t_yankee *x, t_object *attr, long *argc, t_atom **argv)
    {
        char		alloc ;
		double		k = x->kappa ; 
 
        atom_alloc		(argc, argv, &alloc) ;     
        atom_setfloat	(*argv, k) ;
		
        return MAX_ERR_NONE ;
    }
		
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
/*
void check_hashfun (t_hashtab_entry *e, t_yankee *x) 
	{
		t_noeud *nptr = NULL ;
		
		if (e->key) 
			{
				if (nptr = (t_noeud *)e->value)
					{
						post ("CHECK--------- %ld", (long)e->key) ;
						post ("err : %.2lf", nptr->erreur) ;
						post ("utile : %.2lf", nptr->utile) ;
						
						if (nptr->list_arcs)
							{
								post ("nb arcs : %ld", linklist_getsize (nptr->list_arcs)) ;
							}
						
						if (nptr->list_arcs)
							{
								linklist_funall (nptr->list_arcs, (method)check_linkfun, x) ;
							}
					}
			}
	}
	
void check_linkfun (t_object *p, t_yankee *x)
	{
		t_arc *aptr = NULL ;
		
		if (aptr = (t_arc *)p)
			{	
				post ("    dest : %ld / age : %ld", aptr->destination, aptr->age) ;
			}
	}
*/
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------------------------:x